package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.pwhash.*
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import net.sergeych.bipack.Unsigned
import net.sergeych.crypto2.KDF.Argon.Alg.V2id_13
import net.sergeych.crypto2.KDF.Complexity.*
import net.sergeych.crypto2.PBKD.Params

@Serializable
sealed class KDF {

    /**
     * KDF complexity: how many resources (CPY and RAM) it requires to transform a string to a key.
     *
     * Constants started with `Fixed...`, e.g., [FixedLow], [FixedMedium] and [FixedHigher]
     * are first clas constants not to be changed in future but can become obsolete (too simple) with the
     * development of the computer hardware. Other constants can change in future releases, so it is critical
     * to serialize [KDF] or [PBKD.Params] generated with such constants and restore them before
     * deriving keys.
     */
    enum class Complexity {
        FixedLow,
        Interactive,
        FixedMedium,
        Moderate,
        FixedHigher,
        Sensitive,
    }

    abstract fun derive(password: String): UByteArray

    fun deriveMultiple(password: String, count: Int): List<SymmetricKey> {
        val bytes = derive(password)
        val ks = SymmetricKey.keyLength
        check(ks * count <= bytes.size) { "KDF is too short for $count keys: ${bytes.size} we need ${ks * count}" }
        return (0..<count).map {
            val base = it * ks
            Params(this, base, ks).deriveKey(password)
        }
    }

    @Serializable
    @SerialName("argon")
    data class Argon(
        val algorithm: Alg,
        @Unsigned
        val instructionsComplexity: ULong,
        @Unsigned
        val memComplexity: Int,
        val salt: UByteArray,
        @Unsigned
        val keySize: Int,
    ) : KDF(), Comparable<Argon> {

        /**
         * Very abstract strength comparison. If a1 > a2 it generally means that its complexity, and, hopefully,
         * strength is higher.
         */
        override fun compareTo(other: Argon): Int {
            var d = algorithm.ordinal.compareTo(other.algorithm.ordinal)
            if (d != 0) return d
            d = instructionsComplexity.compareTo(other.instructionsComplexity)
            if (d != 0) return d
            d = memComplexity.compareTo(other.memComplexity)
            if (d != 0) return d
            d = keySize
            if (d != 0) return d
            return 0
        }

        enum class Alg(val code: Int) {
            @Suppress("EnumEntryName", "unused")
            V2i_13(crypto_pwhash_argon2i_ALG_ARGON2I13),
            V2id_13(crypto_pwhash_argon2id_ALG_ARGON2ID13),
            ;

            companion object {
                val default = V2id_13
            }
        }

        override fun derive(password: String): UByteArray =
            PasswordHash.pwhash(keySize, password, salt, instructionsComplexity, memComplexity, algorithm.code)

        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other !is Argon) return false

            if (algorithm != other.algorithm) return false
            if (instructionsComplexity != other.instructionsComplexity) return false
            if (memComplexity != other.memComplexity) return false
            if (keySize != other.keySize) return false
            return salt contentEquals other.salt
        }

        override fun hashCode(): Int {
            var result = algorithm.hashCode()
            result = 31 * result + instructionsComplexity.hashCode()
            result = 31 * result + memComplexity
            result = 31 * result + salt.contentHashCode()
            result = 31 * result + keySize
            return result
        }


        companion object {
            val saltSize: Int = crypto_pwhash_SALTBYTES
            val minKeySize: Int = crypto_pwhash_BYTES_MIN.toInt()

            fun randomSalt() = randomUBytes(saltSize)

            fun create(complexity: Complexity, salt: UByteArray, keySize: Int): Argon {
                require(salt.size == saltSize) { "The salt size should be $saltSize" }
                require(keySize > minKeySize) { "The key size should be at least $keySize bytes" }
                return when (complexity) {
                    FixedLow -> Argon(
                        V2id_13,
                        2UL,
                        67108864,
                        salt, keySize
                    )

                    Interactive -> Argon(
                        Alg.default,
                        crypto_pwhash_OPSLIMIT_INTERACTIVE.toULong(),
                        crypto_pwhash_MEMLIMIT_INTERACTIVE,
                        salt, keySize
                    )

                    FixedMedium -> Argon(
                        V2id_13,
                        3UL,
                        268435456,
                        salt, keySize
                    )
                    Moderate -> Argon(
                        Alg.default,
                        crypto_pwhash_OPSLIMIT_MODERATE,
                        crypto_pwhash_MEMLIMIT_MODERATE,
                        salt, keySize
                    )

                    Complexity.Sensitive -> Argon(
                        Alg.default,
                        crypto_pwhash_OPSLIMIT_SENSITIVE,
                        crypto_pwhash_MEMLIMIT_SENSITIVE,
                        salt, keySize
                    )

                    Complexity.FixedHigher -> Argon(
                        V2id_13,
                        4UL,
                        1073741824,
                        salt, keySize
                    )

                }
            }
        }
    }

    companion object {
        @Suppress("unused")
        fun creteDefault(keySize: Int, complexity: Complexity, salt: UByteArray = Argon.randomSalt()): KDF {
            return Argon.create(complexity, salt, keySize)
        }
    }

    data class Instance(val kdf: KDF, val password: String)

}

